/** libfbm, http://www.mare.ee/indrek/libfbm/
  * Copyright (c) 2013, Indrek Mandre <indrek(at)mare.ee>
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions are met: 
  *
  * a Redistributions of source code must retain the above copyright notice, this
  *   list of conditions and the following disclaimer. 
  * b Redistributions in binary form must reproduce the above copyright notice,
  *   this list of conditions and the following disclaimer in the documentation
  *   and/or other materials provided with the distribution. 
  *
  * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
  * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
  * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
  * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
  * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
  * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
  * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
  * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  */

#include <stdio.h>
#include <stddef.h>
#include <assert.h>
#include <math.h>
#include <stdint.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <string.h>

#include <sstream>

#include <gsl/gsl_rng.h>
#include <gsl/gsl_fft_complex.h>

#include "libfbm.hpp"

static bool isPowerOfTwo(size_t v)
{
  if ( !v )
      return false;
  while(1)
    {
      if ( v & 1 )
          return !(v >> 1);
      v = v >> 1;
    }
}

static std::string zvec2str(const libfbm::zvec& v)
{
  std::stringstream ss;
  for ( size_t i = 0; i < v.size(); i++ )
    {
      if ( i )
          ss << "x";
      ss << v[i];
    }
  return ss.str();
}

// not efficient, but this is only called couple of times
static std::string dbl2str(double dbl)
{
  char buf[256];
  sprintf(buf, "%g", dbl);
  return buf;
}

// N-dimensional fast fourier transform
static void FFTND(std::vector<double>& D, const libfbm::zvec& dim, bool forward)
{
  gsl_fft_complex_wavetable *wavetable = 0;
  gsl_fft_complex_workspace *workspace = 0;
  libfbm::zvec itp(dim.size()); // using this to cycle through the N-1 indices
  size_t r = 1; // separation distance of data lements along a column in the data vector D
  std::vector<double> buf; // buffer for the column, stride in gsl fft is 10x slower

  // we run 1D FFT through all columns (with fixed N-1 dimensions)
  // so d is the dimension the column spans
  for ( size_t d = 0; d < dim.size(); d++ )
    {
      bool pt = isPowerOfTwo(dim[d]);

      if ( !pt )
        {
          wavetable = gsl_fft_complex_wavetable_alloc(dim[d]);
          workspace = gsl_fft_complex_workspace_alloc(dim[d]);
        }

      itp.zero();

      while(1)
        {
          // find first element address inside data
          double *dp = &D[2 * itp.index(dim)];
          double *wp = dp;

          if ( d > 0 ) // need to fetch the data into a buffer
            {
              buf.clear();
              buf.reserve(2 * dim[d]);
              for ( int i = 0; i < dim[d]; i++ )
                {
                  buf.push_back(*(dp + 2 * i * r));
                  buf.push_back(*(dp + 2 * i * r + 1));
                }
              wp = &buf[0];
            }

#if 0
          printf("fft(");
          for ( int i = 0; i < dim[d]; i++ )
            {
              if ( i ) printf(", ");
              printf("%.3g", wp[i * 2]);
            }
          printf(")\n");
#endif

          if ( pt )
              gsl_fft_complex_radix2_transform(wp, 1, dim[d], forward ? gsl_fft_forward : gsl_fft_backward);
          else
              gsl_fft_complex_transform(wp, 1, dim[d], wavetable, workspace, forward ? gsl_fft_forward : gsl_fft_backward);

          if ( d > 0 )
            {
              buf.clear();
              buf.reserve(2 * dim[d]);
              for ( int i = 0; i < dim[d]; i++ )
                {
                  *(dp + 2 * i * r) = buf[2 * i];
                  *(dp + 2 * i * r + 1) = buf[2 * i + 1];
                }
            }

          // cycle through all combinations in other dimensions
          if ( !itp.increment_but(dim, d) )
              break;
        }

      if ( !pt )
        {
          gsl_fft_complex_workspace_free(workspace);
          gsl_fft_complex_wavetable_free(wavetable);
        }

      r *= dim[d];
    }
}

libfbm::RandomGenerator::~RandomGenerator()
{
}

libfbm::UniformRandomGenerator::UniformRandomGenerator()
{
  rng = gsl_rng_alloc(gsl_rng_mt19937);
  setSeed(0);
}

libfbm::UniformRandomGenerator::UniformRandomGenerator(unsigned long int seed, const gsl_rng_type* type)
{
  rng = gsl_rng_alloc(type);
  setSeed(seed);
}

libfbm::UniformRandomGenerator::UniformRandomGenerator(const UniformRandomGenerator& copy)
{
  rng = gsl_rng_clone(copy.rng);
}

libfbm::UniformRandomGenerator::~UniformRandomGenerator()
{
  gsl_rng_free(rng);
}

const libfbm::UniformRandomGenerator& libfbm::UniformRandomGenerator::operator=(const libfbm::UniformRandomGenerator& copy)
{
  gsl_rng_free(rng);
  rng = gsl_rng_clone(copy.rng);
  return *this;
}

void libfbm::UniformRandomGenerator::setSeed(unsigned long int seed)
{
  gsl_rng_set(rng, seed);
}

double libfbm::UniformRandomGenerator::next()
{
  return gsl_rng_uniform(rng);
}

libfbm::RandomGenerator *libfbm::UniformRandomGenerator::clone() const
{
  return new UniformRandomGenerator(*this);
}

#define UNLIKELY_GAUSSIAN_VALUE 9e99

libfbm::GaussianRandomGenerator::GaussianRandomGenerator() : urg(new UniformRandomGenerator()), gval2(UNLIKELY_GAUSSIAN_VALUE)
{
}

libfbm::GaussianRandomGenerator::GaussianRandomGenerator(const RandomGenerator *urg) : urg(urg->clone()), gval2(UNLIKELY_GAUSSIAN_VALUE)
{
}

libfbm::GaussianRandomGenerator::GaussianRandomGenerator(const GaussianRandomGenerator& copy) : urg(copy.urg->clone()), gval2(copy.gval2)
{
}

libfbm::GaussianRandomGenerator::~GaussianRandomGenerator()
{
  delete urg;
}

const libfbm::GaussianRandomGenerator& libfbm::GaussianRandomGenerator::operator=(const libfbm::GaussianRandomGenerator& copy)
{
  delete urg;
  gval2 = copy.gval2;
  urg = copy.urg->clone();
  return *this;
}

void libfbm::GaussianRandomGenerator::setSeed(unsigned long int seed)
{
  urg->setSeed(seed);
  gval2 = UNLIKELY_GAUSSIAN_VALUE;
}

double libfbm::GaussianRandomGenerator::next()
{
  if ( gval2 != UNLIKELY_GAUSSIAN_VALUE )
    {
      double ret = gval2;
      gval2 = UNLIKELY_GAUSSIAN_VALUE;
      return ret;
    }

  // Marsaglia polar method
  double u, v, s;

  do {
      u = -1 + 2 * urg->next();
      v = -1 + 2 * urg->next();

      s = u * u + v * v;
  } while ( s >= 1.0 || s == 0 );

  double sr = sqrt(-2.0 * log(s) / s);
  gval2 = u * sr;

  return v * sr;
}

libfbm::RandomGenerator *libfbm::GaussianRandomGenerator::clone() const
{
  return new GaussianRandomGenerator(*this);
}

libfbm::_Cache::_Cache(const std::string& cacheDir, const std::string& cacheName, size_t cacheSize)
  : fp(0), cacheDir(cacheDir), cacheName(cacheName), cacheSize(cacheSize)
{
}

bool libfbm::_Cache::load()
{
  if ( cacheDir.empty() )
      return false;

  char path[1024];
  sprintf(path, "%s/%s.%zu.dat", cacheDir.c_str(), cacheName.c_str(), cacheSize);

  struct stat st;
  mkdir (cacheDir.c_str(), 0700);
  assert(lstat(cacheDir.c_str(), &st) == 0 && "unable to create cache directory");
  assert(st.st_uid == geteuid() && "cache directory not owned by us"); // security check against malicious attacker

  fp = fopen(path, "rb");
  if ( fp )
    {
      fseeko(fp, 0, SEEK_END);
      if ( ftello(fp) == (off_t)(cacheSize * sizeof(double)) )
          return true;
      fclose(fp);
      fp = 0;
    }

  return false;
}

void libfbm::_Cache::store(const std::vector<double>& data)
{
  assert(cacheSize == data.size());

  if ( cacheDir.empty() )
    {
      this->data = data;
      return;
    }

  char path[1024];
  sprintf(path, "%s/%s.%zu.dat", cacheDir.c_str(), cacheName.c_str(), cacheSize);

  struct stat st;
  mkdir (cacheDir.c_str(), 0700);
  assert(lstat(cacheDir.c_str(), &st) == 0 && "unable to create cache directory");
  assert(st.st_uid == geteuid() && "cache directory not owned by us"); // security check against malicious attacker

  fp = fopen(path, "wb");
  assert(fp && "unable to open the cache file for writing");
  fwrite(&data[0], data.size() * sizeof(double), 1, fp);
  fclose(fp);

  fp = fopen(path, "rb");
  assert(fp && "unable to open the cache file for reading");
}

libfbm::_Cache::~_Cache()
{
  if ( fp )
      fclose(fp);
}

libfbm::_CacheReader::_CacheReader(_Cache& cache, size_t bufferSize) : cache(cache)
{
  if ( cache.inMemory() ) // memory-based cache
    {
      bufpos = 0;
      data = &cache.data[0];
    }
  else // file-based cache
    {
      if ( bufferSize > cache.cacheSize )
          bufferSize = cache.cacheSize;
      bufpos = (size_t)-1;
      buf.resize(bufferSize);
      data = &buf[0];
    }
}

void libfbm::_CacheReader::load(size_t i)
{
  bufpos = i - i % buf.size();
  //fprintf(stderr, "%p: load: [%zu, %zu)\n", this, bufpos, buf.size());
  assert(!buf.empty());
  assert(fseeko(cache.fp, bufpos * sizeof(double), SEEK_SET) == 0);
  fread(&buf[0], sizeof(double) * buf.size(), 1, cache.fp);
}

libfbm::SGPContext::SGPContext(const zvec& fieldDim, const zvec& userDim, const std::string& cacheName)
  : fieldDim(fieldDim), userDim(userDim), cache(0), cacheName(cacheName + "." + zvec2str(fieldDim))
{
  assert(fieldDim.size() <= LIBFBM_MAX_DIM);
  rng = new GaussianRandomGenerator();
  bec = 0;
  scaleResult = 1;
}

libfbm::SGPContext::~SGPContext()
{
  delete cache;
  delete rng;
}

void libfbm::SGPContext::postProcess(Field& field)
{
}

void libfbm::SGPContext::setRandomGenerator(const RandomGenerator& newRng)
{
  delete rng;
  rng = newRng.clone();
}

void libfbm::SGPContext::setCacheDir(const std::string& cacheDir)
{
  this->cacheDir = cacheDir;
}

void libfbm::SGPContext::initCache(bool forceRecalc)
{
  if ( cache && forceRecalc )
    {
      delete cache;
      cache = 0;
    }

  if ( cache )
      return;

  size_t len = 1;
  zvec ddim(fieldDim.size());
  for ( size_t i = 0; i < fieldDim.size(); i++ )
    {
      assert(fieldDim[i] >= 2);
      ddim[i] = fieldDim[i] * 2;
      len *= 2 * fieldDim[i];
    }

  cache = new _Cache(cacheDir, cacheName, len);
  if ( !forceRecalc && cache->load() )
      return;

  std::vector<double> IL;
  IL.resize(2 * len, 0);

  std::vector<zvec> pcov;
  double *wp = &IL[0];

  zvec p(fieldDim.size());
  p.zero();
  while(1)
    {
#if 0
      for ( size_t i = 0; i < p.size(); i++ )
          printf(" %d", p[i]);
      printf("\n");
#endif
      pcov.clear();
      pcov.push_back(p);
      for ( size_t i = 0; i < p.size(); i++ )
        {
          int newval1, newval2;

          if ( p[i] < fieldDim[i] )
            {
              newval1 = newval2 = p[i];
            }
          else if ( p[i] > fieldDim[i] )
            {
              newval1 = newval2 = p[i] - 2 * fieldDim[i];
            }
          else
            {
              newval1 = fieldDim[i] - 1;
              newval2 = fieldDim[i] + 1 - 2 * fieldDim[i];
            }

          size_t clen = pcov.size();
          for ( size_t j = 0; j < clen; j++ )
            {
              pcov[j][i] = newval1;
              if ( newval1 != newval2 )
                {
                  pcov.push_back(pcov[j]);
                  pcov.back()[i] = newval2;
                }
            }
        }
      double val = 0;
      for ( size_t i = 0; i < pcov.size(); i++ )
          val += cov(pcov[i]);
      *wp++ = val / pcov.size();
      wp++;
      if ( !p.increment(ddim) )
          break;
    }

#if 0
  for ( size_t i = 0; i < IL.size() / 2; i++ )
    {
      printf ("   %1.5f", IL[2 * i]);
      if ( i && !((i + 1) % 8) )
          printf("\n");
    }
#endif

  FFTND(IL, ddim, false);

  bec = 0;
  double ttrace = 0;
  double ptrace = 0;
  for ( size_t i = 0; i < IL.size(); i += 2 )
    {
      ttrace += IL[i];
      if ( IL[i] < 0 )
        { 
          if ( IL[i] < -1e-14 ) // allow some numeric errors
            {
              bec++;
              //printf("Neg: %g\n", IL[i]);
            }
          IL[i] = 0;
        }
      ptrace += IL[i];
      IL[i/2] = sqrt(IL[i] / len) * scaleResult;
    }
  IL.resize(IL.size() / 2);
  if ( bec )
    {
      fprintf(stderr, "Zeroed %zu/%zu of negative eigenvalues (matrix not nonnegative definite, generated noise will be approximate)\n", bec, IL.size());
      double mul = sqrt(ttrace / ptrace);
      for ( size_t i = 0; i < IL.size(); i++ )
          IL[i] *= mul;
    }

#if 0
  printf("Results:\n");
  for ( size_t i = 0; i < IL.size(); i++ )
    {
      printf ("   %1.5f", IL[i]);
      if ( i && !((i + 1) % 8) )
          printf("\n");
    }
#endif

  cache->store(IL);
}

libfbm::Field::Field(SGPContext& context, bool allowCorrelated)
  : context(context), allowCorrelated(allowCorrelated), bufferSize(65536),
    psel(!allowCorrelated ? 1 : (context.getFieldDim().size() + 1))
{
  memset(muls, 0, sizeof(muls));
  size_t r = 1;
  for ( size_t i = 0; i < context.getFieldDim().size(); i++ )
    {
      muls[i] = 2 * r;
      r = r * 2 * context.getFieldDim()[i];
    }
}

libfbm::Field::~Field()
{
}

void libfbm::Field::setBufferSize(size_t bufferSize)
{
  this->bufferSize = bufferSize;
}

void libfbm::Field::clear()
{
  std::vector<double> cl;
  Z.swap(cl);
}

void libfbm::Field::generate()
{
  zvec ddim = context.getFieldDim() * 2;

  if ( !Z.empty() && psel.increment(2) )
    {
      size_t offset = 0;
      if ( psel[0] )
          offset++;
      if ( allowCorrelated )
        {
          size_t r = 1;
          for ( size_t i = 1; i < psel.size(); i++ )
            {
              if ( psel[i] )
                  offset += 2 * r * context.getFieldDim()[i - 1];
              r *= ddim[i - 1];
            }
        }
      datap = &Z[offset];
      context.postProcess(*this);
      return;
    }

  if ( !context.cache )
      context.initCache();
  _CacheReader LF(*context.cache, bufferSize);
  Z.resize(2 * context.cache->size());
  datap = &Z[0];
  psel.zero();

  for ( size_t i = 0; i < Z.size() / 2; i++ )
    {
      Z[i * 2] = LF[i] * context.rng->next();
      Z[i * 2 + 1] = LF[i] * context.rng->next();
    }

  FFTND(Z, ddim, true);
  context.postProcess(*this);
}

void libfbm::Field::integrate()
{
  // we must basically cumulatively sum all columns in all directions
  // so just like in the FFTND, but instead of fft we sum
  zvec dim = context.getFieldDim();

  libfbm::zvec itp(dim.size()); // using this to cycle through the N-1 indices

  for ( size_t d = 0; d < dim.size(); d++ )
    {
      itp.zero();
      size_t r = muls[d];

      while(1)
        {
          // find first element address inside data
          double *dp = &at(itp);

          double sum = 0;
          for ( int i = 0; i < dim[d]; i++ )
            {
              sum += dp[i * r];
              dp[i * r] = sum;
            }

          // cycle through all combinations in other dimensions
          if ( !itp.increment_but(dim, d) )
              break;
        }
    }
}

libfbm::FGNContext::FGNContext(double H, const zvec& dim)
  : SGPContext(dim, dim, "fgn." + dbl2str(H)), H(H)
{
}

double libfbm::FGNContext::cov(const zvec& p)
{
  double ret = 1;
  for ( size_t i = 0; i < p.size(); i++ )
    {
      double r = fabs(p[i]);
      ret *= 0.5 * (pow((r - 1) * (r - 1), H) - 2 * pow(r * r, H) + pow((r + 1) * (r + 1), H));
    }
  return ret;
}

void libfbm::FWSContext::postProcess(Field& field)
{
  field.integrate();
}

static libfbm::zvec dim2dim(size_t size, size_t dim)
{
  assert(dim <= LIBFBM_MAX_DIM);
  libfbm::zvec ret(dim);
  for ( size_t i = 0; i < dim; i++ )
      ret[i] = size;
  return ret;
}

#define CONSERVATIVE_R 1.3

//  {4096, {{0.775, 1.00079}, {0.8, 1.00108}, {0.825, 1.0084}, {0.85, 1.02536}, {0.875, 1.05198}, {0.9, 1.08649}, {0.925, 1.12773}, {0.95, 1.17189}, {0.975, 1.21928}, {0.99, 1.26256}}},
static double R_map2[][2] = {
  {0.775, 1.001},
  {0.8,   1.002},
  {0.825, 1.009},
  {0.85,  1.026},
  {0.875, 1.052},
  {0.9,   1.087},
  {0.925, 1.128},
  {0.95,  1.172},
  {0.975, 1.220},
  {0.99,  1.263},
  {1,     CONSERVATIVE_R},
  {0,     0}
};

size_t libfbm::FBMSteinContext::userSize2FieldSize(size_t size, double R)
{
  size = ceil(R * size * sqrt(2) + 1e-16);
  while ( size % 8 )
      size++; // get some even-ness to optimize the FFT
  return size;
}

double libfbm::FBMSteinContext::getRForH(double H, double R)
{
  if ( H <= 0.75 )
      return 1.0;

  if ( R < 1 )
    {
      size_t i = 0;
      while ( R_map2[i][0] )
        {
          if ( H <= R_map2[i][0] )
            {
              R = R_map2[i][1];
              //printf("Found: %g\n", R);
              break;
            }
          i++;
        }
      if ( R < 1 )
          R = CONSERVATIVE_R;
    }

  return R;
}

size_t libfbm::FBMSteinContext::fieldSize2UserSize(size_t size, double R)
{
  return floor(size / (R * sqrt(2)) - 1e-16);
}

// This is kind of insane. Having the dimension in the constructor actually
// makes this really cumbersome and ugly. It's convenient for the user. Oh well.
libfbm::FBMSteinContext::FBMSteinContext(double H, size_t dim, size_t size, double Rhint, bool mapDim)
  : SGPContext(
      dim2dim(mapDim ? userSize2FieldSize(size, getRForH(H, Rhint)) : size, dim),
      dim2dim(mapDim ? size : fieldSize2UserSize(size, getRForH(H, Rhint)), dim),
      "stein." + dbl2str(H) + "." + dbl2str(getRForH(H, Rhint))),
    H(H), R(getRForH(H, Rhint)), pp(true)
{
  if ( H <= 0.75 )
    {
      R = 1;
      beta = 0;
      c2 = H;
      c0 = 1 - c2;
    }
  else
    {
      if ( R < 1 )
          R = CONSERVATIVE_R;
      beta = 2 * H * (2 - 2 * H) / (3 * R * (R * R - 1));
      c2 = (2 * H - beta * (R - 1) * (R - 1) * (R + 2)) / 2;
      c0 = beta * (R - 1) * (R - 1) * (R - 1) + 1 - c2;
    }
  Rscale = R / (getFieldDim()[0] - 1) + 1e-16;
  //assert(floor((sqrt(2) / 2) * (1 / Rscale) - 1e-16) >= dim[0] - 1);

  setScaleResult(0.5 * sqrt(2) * pow(Rscale, -H));
}

double libfbm::FBMSteinContext::cov(const zvec& p)
{
  uint64_t len2 = 0;
  for ( size_t i = 0; i < p.size(); i++ )
      len2 += (uint64_t)p[i] * p[i];
  double len = sqrt(len2) * Rscale;
  double ret;
  if ( len <= 1 )
    {
      ret = c0 + c2 * len * len - pow(len, 2 * H);
    }
  else if ( len <= R )
    {
      ret = beta * (R - len) * (R - len) * (R - len) / len;
    }
  else
    {
      ret = 0;
    }
  //printf("cov(%d,%d) len=%g ret=%g\n", p[0], p[1], len, ret);
  return ret;
}

void libfbm::FBMSteinContext::postProcess(Field& field)
{
  if ( !pp )
      return;
  double scalingCorrector = 0.5 * sqrt(2) * pow(Rscale, -H);

  int range = getDim()[0];
  const zvec& dim = getFieldDim();

  std::vector<double> Z;
  Z.reserve(dim.size());
  for ( size_t i = 0; i < dim.size(); i++ )
      Z.push_back(sqrt(2 * c2) * nextRng());

  std::vector<double> r;
  r.reserve(dim.size() * range);
  for ( size_t d = 0; d < dim.size(); d++ )
      for ( int i = 0; i < range; i++ )
          r.push_back(i * Rscale * Z[d] * scalingCorrector);

  zvec p(dim.size());
  p.zero();
  double corner = field(p);

  while (1)
    {
      double sum = -corner;
      for ( size_t i = 0; i < p.size(); i++ )
          sum += r[i * range + p[i]];
      double& v = field.at(p);
      v = (v + sum);
      if ( !p.increment(range) )
          break;
    }
}

void libfbm::FBMSteinContext::disablePostProcessing()
{
  pp = false;
}

void libfbm::SGPTester::test(SGPContext& context, size_t printInterval)
{
  Field fbm(context, true);

  const zvec& dim = context.getFieldDim();
  size_t len = 1;
  for ( size_t i = 0; i < dim.size(); i++ )
      len *= dim[i];
  size_t totE = 2 * ((size_t)1 << dim.size());

  std::vector<long double> sums(len * len * totE, 0);

  size_t count = 0;
  while(1)
    {
      size_t index = 0;
      zvec v1(dim.size());
      zvec v2(dim.size());
      for ( size_t E = 0; E < totE; E++ )
        {
          fbm.generate();

          v1.zero();
          while(1)
            {
              v2.zero();
              while(1)
                {
                  sums[index++] += fbm(v1) * fbm(v2);
                  if ( !v2.increment(dim) )
                      break;
                }
              if ( !v1.increment(dim) )
                  break;
            }
        }
      count++;

      if ( !(count % printInterval) )
        {
          printf("%zu\n", count);
          size_t index = 0;
          zvec v1(dim.size());
          zvec v2(dim.size());
          for ( size_t E = 0; E < totE; E++ )
            {
              double maxabserr = 0, maxrelerr = 0;
              double maxrelerrcov = 0, maxrelerrv = 0, maxabserrv = 0, maxabserrcov = 0;
              v1.zero();
              while(1)
                {
                  v2.zero();
                  while(1)
                    {
                      // cov of v1 versus v2
                      zvec dif(dim.size());
                      for ( size_t i = 0; i < dim.size(); i++ )
                          dif[i] = v1[i] - v2[i];
                      double cov = context.cov(dif);
                      double v = sums[index++] / count;

                      if ( maxabserr < fabs(v - cov) )
                        {
                          maxabserr = fabs(v - cov);
                          maxabserrv = v;
                          maxabserrcov = cov;
                        }
                      if ( fabs(cov) > 1e-15 && maxrelerr < fabs((v - cov)/cov) )
                        {
                          maxrelerr = fabs((v - cov)/cov);
                          maxrelerrv = v;
                          maxrelerrcov = cov;
                        }

                      if ( !v2.increment(dim) )
                          break;
                    }
                  if ( !v1.increment(dim) )
                      break;
                }
              printf("%zu: maxabserr=%g (%g vs cov %g), maxrelerr=%g (%g vs cov %g)\n",
                  E,
                  maxabserr, maxabserrv, maxabserrcov,
                  maxrelerr, maxrelerrv, maxrelerrcov);
            }
        }
    }
}

